/**********************************************************************************
 Copyright (C) 2010
 by 52 North Initiative for Geospatial Open Source Software GmbH

 Contact: Andreas Wytzisk 
 52 North Initiative for Geospatial Open Source Software GmbH
 Martin-Luther-King-Weg 24
 48155 Muenster, Germany
 info@52north.org

 This program is free software; you can redistribute and/or modify it under the
 terms of the GNU General Public License version 2 as published by the Free
 Software Foundation.

 This program is distributed WITHOUT ANY WARRANTY; even without the implied
 WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License along with this 
 program (see gnu-gplv2.txt). If not, write to the Free Software Foundation, Inc., 
 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or visit the Free Software
 Foundation web page, http://www.fsf.org.
 
 Created on: 02.11.2010
 *********************************************************************************/

package org.n52.spsMockup;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.opengis.sps.x10.GetCapabilitiesDocument;
import net.opengis.sps.x10.GetCapabilitiesDocument.GetCapabilities;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.n52.spsMockup.OwsExceptionReport.ExceptionCode;

/**
 * 
 * @author Daniel Nüst (daniel.nuest@uni-muenster.de)
 * 
 */
public class SpsMockup extends HttpServlet {

    /**
     * 
     */
    private static final long serialVersionUID = 2433606218994581512L;

    private static final String GET_REQUEST_PARAM = "request";

    private static final String GET_CAPABILITIES_NAME = "GetCapabilities";

    private static final String SERVICE_PARAM = "service";

    private static final String SPS_SERVICE_NAME = "SPS";

    private static final String SENSOR_DESCRIPTION_PARAM = "sensor";

    private static final String RESPONSES_PATH = "WEB-INF/responses/";

    private static final String IFGICAM_FILE = "ifgicam-discoveryprofile.xml";

    private static final String IFGICAM_ID = "urn:ogc:object:feature:Sensor:52n:ifgicam_flur";

    private static final String CAPABILITIES_FILE = "capabilities.xml";

    private static final String SENSOR_DESCRIPTION_REPLACEMENT_PATTERN = "@@ifgicam_definition_url@@";

    /**
     * 
     */
    private static final String SERVICE_URL = "http://giv-genesis.uni-muenster.de:8080/SpsMockup/sps";
//    private static final String SERVICE_URL = "http://localhost:8080/SpsMockup/sps";

    private static Logger log;

    private SpsMockupResponse capsResponse;

    private SpsMockupResponse ifgicamResponse;

    /*
     * (non-Javadoc)
     * 
     * @see javax.servlet.GenericServlet#init()
     */
    @Override
    public void init() throws ServletException {
        super.init();

        ServletContext context = getServletContext();
        String responsesPath = context.getRealPath("/") + RESPONSES_PATH;

        // initialize Logger
        PropertyConfigurator.configureAndWatch("log4j.xml", 60 * 1000);
        log = Logger.getLogger(SpsMockup.class);

        // initialize responses
        try {
            File f = new File(responsesPath + IFGICAM_FILE);
            this.ifgicamResponse = new SpsMockupResponse(XmlObject.Factory.parse(f));

            f = new File(responsesPath + CAPABILITIES_FILE);
            XmlObject tempXml = XmlObject.Factory.parse(f);
            String pattern = SENSOR_DESCRIPTION_REPLACEMENT_PATTERN;
            String replace = SERVICE_URL + "?" + SENSOR_DESCRIPTION_PARAM + "=" + IFGICAM_ID;
            String tempXMLString = tempXml.xmlText();
            tempXMLString = tempXMLString.replace(pattern, replace);

            this.capsResponse = new SpsMockupResponse(XmlObject.Factory.parse(tempXMLString));
        }
        catch (XmlException e) {
            log.error("Could NOT initialize SPS Mockup Service!", e);
            return;
        }
        catch (IOException e) {
            log.error("Could NOT initialize SPS Mockup Service!", e);
            return;
        }

        log.info("** INITIALIZED SPS Mockup Service **");
    }

    @Override
    public void destroy() {
        super.destroy();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest ,
     * javax.servlet.http.HttpServletResponse)
     */
    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        log.debug(" ****** (GET) Connected from: " + req.getRemoteAddr() + " " + req.getRemoteHost());
        String queryString = req.getQueryString();
        if (log.isDebugEnabled())
            log.debug("Query String: " + queryString);

        SpsMockupResponse spsResp = doGetOperation(queryString);
        doResponse(resp, spsResp);
    }

    /**
     * @param queryString
     * @return
     * @throws OwsExceptionReport
     */
    private SpsMockupResponse doGetOperation(String queryString) {
        // check queryString
        if (queryString != null && queryString.length() != 0) {
            String[] params = queryString.split("&");

            // just check for service and getcapabilities, ignore other parameters
            boolean foundGetCap = false;
            boolean foundSps = false;
            for (String param : params) {
                String[] nameAndValue = param.split("=");
                // check the request parameter
                if (nameAndValue[0].equalsIgnoreCase(GET_REQUEST_PARAM)) {
                    // check request = GetCapabilities
                    if (nameAndValue[1].equalsIgnoreCase(GET_CAPABILITIES_NAME)) {
                        foundGetCap = true;
                    }
                }
                else if (nameAndValue[0].equalsIgnoreCase(SERVICE_PARAM)) {
                    // check request = GetCapabilities
                    if (nameAndValue[1].equalsIgnoreCase(SPS_SERVICE_NAME)) {
                        foundSps = true;
                    }
                }
                else if (nameAndValue[0].equalsIgnoreCase(SENSOR_DESCRIPTION_PARAM)) {
                    if (log.isDebugEnabled())
                        log.debug("**** Get Sensor Description for " + nameAndValue[1]);
                    if (nameAndValue[1].equalsIgnoreCase(IFGICAM_ID)) {
                        return this.ifgicamResponse;
                    }
                    log.error("Sensor description not found!");
                    OwsExceptionReport se = new OwsExceptionReport(OwsExceptionReport.ExceptionCode.InvalidParameterValue,
                                                                   SENSOR_DESCRIPTION_PARAM,
                                                                   "Sensor description not found!");
                    return new SpsMockupResponse(se.getDocument());
                }
            }

            if (foundGetCap && foundSps) {
                if (log.isDebugEnabled())
                    log.debug("**** GetCapabilities");
                return this.capsResponse;
            }
            if (foundGetCap || foundSps) {
                log.error("Invalid Get request! 2 Parameters needed!");
                OwsExceptionReport se = new OwsExceptionReport(OwsExceptionReport.ExceptionLevel.DetailedExceptions);
                se.addCodedException(OwsExceptionReport.ExceptionCode.InvalidRequest,
                                     "doGetOperation()",
                                     "Invalid Get request! 2 Parameters needed: 'service' and 'request'!");
                return new SpsMockupResponse(se.getDocument());
            }

            log.error("Mockup service, only GetCapabilities is supported!");
            OwsExceptionReport se = new OwsExceptionReport(OwsExceptionReport.ExceptionCode.InvalidRequest,
                                                           null,
                                                           "Mockup service, only GetCapabilities is supported!");
            return new SpsMockupResponse(se.getDocument());

        }
        log.error("Invalid Get request!");
        OwsExceptionReport se = new OwsExceptionReport(OwsExceptionReport.ExceptionLevel.DetailedExceptions);
        se.addCodedException(OwsExceptionReport.ExceptionCode.InvalidRequest,
                             "HttpGetRequestDecoder.receiveRequest()",
                             "Invalid Get request!");
        return new SpsMockupResponse(se.getDocument());
    }

    @Override
    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        if (log.isDebugEnabled())
            log.debug(" ****** (POST) Connected from: " + req.getRemoteAddr() + " " + req.getRemoteHost());

        // Read the request
        InputStream in = req.getInputStream();
        String inputString = "";

        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        String line;
        StringBuffer sb = new StringBuffer();
        while ( (line = br.readLine()) != null) {
            sb.append(line + "\n");
        }
        br.close();
        inputString = sb.toString();

        // discard "request="
        if (inputString.startsWith("request=")) {
            inputString = inputString.substring(8, inputString.length());
            inputString = java.net.URLDecoder.decode(inputString, "UTF-8");
        }

        SpsMockupResponse spsResp = doPostOperation(inputString);
        doResponse(resp, spsResp);
    }

    /**
     * @param inputString
     * @return
     */
    private SpsMockupResponse doPostOperation(String inputString) {
        XmlObject requestDoc;

        try {
            requestDoc = XmlObject.Factory.parse(inputString);
        }
        catch (XmlException e) {
            log.error("Could not parse request.", e);
            return new SpsMockupResponse(new OwsExceptionReport(e).getDocument());
        }

        // support get capabilities
        if (requestDoc instanceof GetCapabilitiesDocument) {
            GetCapabilitiesDocument getCapDoc = (GetCapabilitiesDocument) requestDoc;
            GetCapabilities getCap = getCapDoc.getGetCapabilities();

            String service = getCap.getService();

            if (service.equals(SPS_SERVICE_NAME)) {
                if (log.isDebugEnabled())
                    log.debug("**** GetCapabilities");
                return this.capsResponse;
            }
        }

        return new SpsMockupResponse(new OwsExceptionReport(ExceptionCode.InvalidRequest,
                                                            null,
                                                            "Mockup service, only GetCapabilities is supported!").getDocument());
    }

    /**
     * 
     * @param resp
     * @param sirResp
     */
    private void doResponse(HttpServletResponse resp, SpsMockupResponse sirResp) {
        try {
            String contentType = sirResp.getContentType();
            int contentLength = sirResp.getContentLength();
            byte[] bytes = sirResp.getByteArray();
            resp.setContentLength(contentLength);
            OutputStream out = resp.getOutputStream();
            resp.setContentType(contentType);
            out.write(bytes);
            out.close();
        }
        catch (IOException ioe) {
            log.error("doResponse", ioe);
        }
    }
}